@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2009, 2010, 2011, 2012, 2013 Free Software
+  Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015 Free Software
   Foundation, Inc.
 
   This file is part of GNU Inetutils.
@@ -72,12 +72,6 @@ static int host_family = AF_INET;
 
 
 
-#ifdef LOG_NFACILITIES
-# define IU_MAX_FAC LOG_NFACILITIES
-#else
-# define IU_MAX_FAC 32
-#endif
-
 int
 decode (char *name, CODE *codetab, const char *what)
 {
@@ -88,8 +82,16 @@ decode (char *name, CODE *codetab, const char *what)
       char *p;
       unsigned long n = strtoul (name, &p, 0);
 
-      if (*p || n >= IU_MAX_FAC)	/* Includes range errors.  */
-	error (EXIT_FAILURE, 0, "%s: invalid %s number", what, name);
+      /* For portability reasons, a numerical facility is entered
+       * as a decimal integer, shifted left by three binary bits.
+       * Any overflow check must adapt to this fact.
+       * For the purpose of remote logging from a local system,
+       * tests based on LOG_NFACILITIES are insufficient, as a
+       * remote system may well distinguish more facilities than
+       * the local system does!
+       */
+      if (*p || n > LOG_FACMASK)	/* Includes range errors.  */
+	error (EXIT_FAILURE, 0, "invalid %s number: %s", what, name);
 
       return n;
     }
@@ -230,15 +232,14 @@ open_socket (void)
 
 	  if (source)
 	    {
-	      /* Make the assumption that the source address
-	       * encodes a desired domain family and that it
-	       * be numeric.  */
+	      /* Resolver uses the same address family
+	       * as the already resolved target host.
+	       */
 	      int ret;
 	      struct addrinfo tips, *a;
 
 	      memset (&tips, 0, sizeof (tips));
 	      tips.ai_family = ai->ai_family;
-	      tips.ai_flags = AI_NUMERICHOST;
 
 	      ret = getaddrinfo(source, NULL, &tips, &a);
 	      if (ret)
